/*
 * Kosmos.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package hu.midori.kosmos.server.cc;

import hu.midori.kosmos.model.CcBuildStatus;
import hu.midori.kosmos.model.CcProject;
import hu.midori.kosmos.model.CcTest;
import hu.midori.kosmos.model.CcTestResult;
import hu.midori.kosmos.model.ZoomableImageLocator;
import hu.midori.kosmos.protocol.CcService;
import hu.midori.kosmos.server.AbstractKosmosService;
import hu.midori.kosmos.server.util.ChartUtils;
import hu.midori.kosmos.server.util.ColorCodedKey;
import hu.midori.kosmos.server.util.WebCrawlingUtils;
import hu.midori.kosmos.server.util.UrlUtils;

import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import net.sf.saxon.om.NodeInfo;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * Implementation of the <i>CruiseControl</i> service.
 * <p>
 * Downloads the log XML embedded in the CC results webpage
 * and analyzes it using running <i>XQuery</i>es.
 *
 * @author <a href="mailto:aron.gombas@midori.hu">Aron Gombas</a>
 * @version $Id$
 */
public class CcServiceImpl extends AbstractKosmosService implements CcService {
    private final static Log log = LogFactory.getLog(CcServiceImpl.class);

    @SuppressWarnings("unchecked")
    public List<CcProject> getProjects(String url) {
        log.debug("Returning projects...");

        List<CcProject> projects = new ArrayList<CcProject>();
        updateTimestamp();

        beginStoreSession();

        try {
            Document dom = WebCrawlingUtils.downloadHtmlDom(new URL(url));

            // scrape "XML Log File" link from the "Build Results" page and download it
            Node aNode = WebCrawlingUtils.findDomNodeByAttribute(dom.getElementsByTagName("a"), "href", "tab=xmlLogFile");
            String xmlLogPageUrl = UrlUtils.getBaseUrl(url) + aNode.getAttributes().getNamedItem("href").getNodeValue();
            log.debug(String.format("Going to XML log page at \"%s\"", xmlLogPageUrl));
            Document xmlLogPageDom = WebCrawlingUtils.downloadHtmlDom(new URL(xmlLogPageUrl));

            // scrape embedded XML log
            Node preNode = WebCrawlingUtils.findDomNodeByAttribute(xmlLogPageDom.getElementsByTagName("pre"), "class", "modifications-data");
            String xmlLog = preNode.getFirstChild().getNodeValue();
            log.debug(String.format("XML log of length %d found", xmlLog.length()));

            // scape "global" properties
            Document xmlLogDom = WebCrawlingUtils.parseStringDom(xmlLog);

            String name = WebCrawlingUtils.runXQueryString(xmlLogDom, "<dummy> { data(//property[@name='projectname']/@value) } </dummy>");
            String buildLabel = WebCrawlingUtils.runXQueryString(xmlLogDom, "<dummy> { data(//property[@name='label']/@value) } </dummy>");
            int status = (WebCrawlingUtils.runXQueryString(xmlLogDom, "<dummy> { data(//build/@error) } </dummy>").length() == 0) ? CcBuildStatus.SUCCEEDED : CcBuildStatus.FAILED;
            Date timestamp = CcConstants.CC_BUILD_DATEFORMAT.parse(WebCrawlingUtils.runXQueryString(xmlLogDom, "<dummy> { data(//property[@name='cctimestamp']/@value) } </dummy>"));
            int modifications = WebCrawlingUtils.runXQueryInt(xmlLogDom, "<dummy> { data(count(//modification)) } </dummy>");
            String buildTime = WebCrawlingUtils.runXQueryString(xmlLogDom, "<dummy> { data(//build/@time) } </dummy>");// TODO if we keep it as String, that cannot be localized -> should be intelligently parsed (using a Duration class maybe?) and stored as Object
            int tests = WebCrawlingUtils.runXQueryInt(xmlLogDom, "<dummy> { data(count(//testcase)) } </dummy>");
            int errors = WebCrawlingUtils.runXQueryInt(xmlLogDom, "<dummy> { data(count(//testcase/error)) } </dummy>");
            int failures = WebCrawlingUtils.runXQueryInt(xmlLogDom, "<dummy> { data(count(//testcase/failure)) } </dummy>");

            // scrape testcases
            List<CcTest> testCases = new ArrayList<CcTest>();
            String query =
                "for " +
                    "$d in //testcase " +
                "return " +
                    "<dummy>" +
                        "{ data($d/@classname) }|" +
                        "{ data($d/@name) }|" +
                        "{ count($d/failure) }|" +
                        "{ count($d/error) }" +
                    "</dummy>";

            List result = WebCrawlingUtils.runXQuery(xmlLogDom, query);
            for(Iterator it = result.iterator(); it.hasNext();) {
                String value = WebCrawlingUtils.eliminateEmptyValues(((NodeInfo)it.next()).getStringValue());
                log.debug(String.format("Scraped testcase: \"%s\"", value));

                StringTokenizer tokenizer = new StringTokenizer(value, "|");
                String className = tokenizer.nextToken();
                String testCaseName = tokenizer.nextToken();
                int result_ = CcTestResult.SUCCEEDED;
                if(!tokenizer.nextToken().equals("0"))
                    result_ = CcTestResult.FAILED;
                else if(!tokenizer.nextToken().equals("0"))
                    result_ = CcTestResult.ERROR;

                testCases.add(new CcTest(result_, className, testCaseName));
            }

            // generate chart
            List<Map.Entry<ColorCodedKey, Integer>> testsPerCategory = new ArrayList<Map.Entry<ColorCodedKey, Integer>>();
            int success = tests - errors - failures;
            if(success > 0)
                testsPerCategory.addAll(Collections.singletonMap(new ColorCodedKey(getResourceString("hu.midori.kosmos.server.cc.cc_monitoring", "ccproject.testsSucceded"), CcConstants.SUCCESS_COLOR), success).entrySet());
            if(failures > 0)
                testsPerCategory.addAll(Collections.singletonMap(new ColorCodedKey(getResourceString("hu.midori.kosmos.server.cc.cc_monitoring", "ccproject.failures"), CcConstants.FAILURE_COLOR), failures).entrySet());
            if(errors > 0)
                testsPerCategory.addAll(Collections.singletonMap(new ColorCodedKey(getResourceString("hu.midori.kosmos.server.cc.cc_monitoring", "ccproject.errors"), CcConstants.ERROR_COLOR), errors).entrySet());

            ZoomableImageLocator testsPerCategoryChartLocator = storeChart(url + "_tests_per_category", ChartUtils.generateColorCodedPieChart(ChartUtils.collectionToPieDataset(testsPerCategory)));

            // save
            CcProject project = new CcProject(name, url, buildLabel, status, timestamp,
                modifications, buildTime,
                tests, errors, failures, trimList(testCases),
                testsPerCategoryChartLocator);
            log.debug(String.format("Adding %s...", project));
            projects.add(project);
        } catch(Exception ex) {
            log.error("Unable to scrape", ex);
        }

        endStoreSession();

        return projects;
    }
}
