/*
 * Kosmos.
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package hu.midori.kosmos.server.jira;

import hu.midori.kosmos.model.JiraProject;
import hu.midori.kosmos.model.ZoomableImageLocator;
import hu.midori.kosmos.protocol.JiraService;
import hu.midori.kosmos.server.util.ChartUtils;
import hu.midori.kosmos.server.util.WebCrawlingUtils;
import hu.midori.kosmos.server.util.UrlUtils;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import net.sf.saxon.om.NodeInfo;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;

/**
 * Implementation of the <i>JIRA</i> service.
 * <p>
 * Downloads JIRA webpages and analyze them using web-crawling techniques.
 * Useful in environments where the JIRA remote API is not enabled.
 *
 * @author <a href="mailto:aron.gombas@midori.hu">Aron Gombas</a>
 * @version $Id$
 */
public class JiraWebCrawlingServiceImpl extends AbstractJiraServiceImpl implements JiraService {
    private final static Log log = LogFactory.getLog(JiraWebCrawlingServiceImpl.class);

    @SuppressWarnings("unchecked")
    public List<JiraProject> getProjects(String url) {
        log.debug("Returning projects...");

        List<JiraProject> projects = new ArrayList<JiraProject>();
        updateTimestamp();

        beginStoreSession();

        try {
            // download database info
            connect(url);

            Document dom = WebCrawlingUtils.downloadHtmlDom(new URL(UrlUtils.removeUserInfoFromUrl(url)));

            disconnect(url);

            // run XQueries
            String queryProject =
                "for " +
                    "$d in //table " +
                "where " +
                    "contains($d/child::tr/child::td/child::h3/child::a/@href, '/secure/BrowseProjects.jspa') " +
                "return " +
                    "<dummy>" +
                        "{ data(fn:substring-after(fn:normalize-space($d/child::tr/child::td/child::h3/text()), ' ')) }|" +
                        "{ data(fn:substring-after(fn:substring-before($d/child::tr/child::td/child::h3/font/text(), ')'), ' ')) }|" +
                        "{ data($d/following-sibling::table/child::tr/child::td/child::p[1]/child::a[2]) }|" +
                        "{ data(fn:normalize-space(fn:string-join($d/following-sibling::table/child::tr/child::td/child::p[1]/text(), ''))) }|" +
                        "{ data($d/following-sibling::table/child::tr/child::td/child::p[1]/child::a[1]) }" +
                    "</dummy>";
            List resultProject = WebCrawlingUtils.runXQuery(dom, queryProject);
            if(resultProject.size() < 1)
                throw new IllegalStateException(String.format("Couldn't find JIRA project properties (%d) at \"%s\"", resultProject.size(), url));

            String queryIssueStatus =
                "for " +
                    "$d in //tr " +
                "where " +
                    "count($d/child::td) = 4 and " +
                    "contains($d/child::td[2]/child::a[1]/@href, '/secure/IssueNavigator.jspa') and " +
                    "contains($d/child::td[2]/child::a[1]/@href, 'status') " + // can be "status" or "statusIds"
                "return " +
                    "<dummy>" +
                        "{ data(fn:substring-after(fn:substring-after($d/child::td[2]/child::a[1]/@href, 'status'), '=')) }|" +// TODO should grab the next number after "status"
                        "{ data(fn:normalize-space(fn:string-join($d/child::td[3]/text(), ''))) }" +
                    "</dummy>";
            List resultIssueStatus = WebCrawlingUtils.runXQuery(dom, queryIssueStatus);

            String queryIssuePriority =
                "for " +
                    "$d in //tr " +
                "where " +
                    "count($d/child::td) = 4 and " +
                    "contains($d/child::td[2]/child::a[1]/@href, '/secure/IssueNavigator.jspa') and " +
                    "contains($d/child::td[2]/child::a[1]/@href, 'priority') " +
                "return " +
                    "<dummy>" +
                        "{ data(fn:substring(fn:substring-after(fn:substring-after($d/child::td[2]/child::a[1]/@href, 'priority'), '='), 1, 1)) }|" +// TODO not appropriate, because priority is hardwired to a single character
                        "{ data(fn:normalize-space(fn:string-join($d/child::td[3]/text(), ''))) }" +
                    "</dummy>";
            List resultIssuePriority = WebCrawlingUtils.runXQuery(dom, queryIssuePriority);

            String queryIssueAssignee =
                "for " +
                    "$d in //tr " +
                "where " +
                    "count($d/child::td) = 3 and " +
                    "contains($d/child::td[1]/child::a[1]/@href, '/secure/IssueNavigator.jspa') and " +
                    "contains($d/child::td[1]/child::a[1]/@href, 'assignee') " +
                "return " +
                    "<dummy>" +
                        "{ data(fn:normalize-space($d/child::td[1]/child::a[1]/text())) }|" +
                        "{ data(fn:normalize-space(fn:string-join($d/child::td[2]/text(), ''))) }" +
                    "</dummy>";
            List resultIssueAssignee = WebCrawlingUtils.runXQuery(dom, queryIssueAssignee);

            // scrape
            for(Iterator itProject = resultProject.iterator(); itProject.hasNext();) {
                String valueProject = WebCrawlingUtils.eliminateEmptyValues(((NodeInfo)itProject.next()).getStringValue());
                log.debug(String.format("Scraped project: \"%s\"", valueProject));

                StringTokenizer tokenizerProject = new StringTokenizer(valueProject, "|");
                String name = tokenizerProject.nextToken();
                String projectKey = tokenizerProject.nextToken();
                String projectUrl = tokenizerProject.nextToken();
                String description = tokenizerProject.nextToken();
                String lead = tokenizerProject.nextToken();

                // scrape issue statuses
                int openIssues = 0;
                int codingInProgressIssues = 0;
                int reopenedIssues = 0;
                int resolvedIssues = 0;
                int closedIssues = 0;
                for(Iterator itIssueStatus = resultIssueStatus.iterator(); itIssueStatus.hasNext();) {
                    String valueIssueStatus = ((NodeInfo)itIssueStatus.next()).getStringValue();
                    log.debug(String.format("Scraped issue status: \"%s\"", valueIssueStatus));

                    StringTokenizer tokenizerIssueStatus = new StringTokenizer(valueIssueStatus, "|");
                    int status = Integer.parseInt(tokenizerIssueStatus.nextToken());
                    int count = Integer.parseInt(tokenizerIssueStatus.nextToken());
                    switch(status) {
                        case 1: openIssues = count; break;
                        // TODO what is 2?
                        case 3: codingInProgressIssues = count; break;
                        case 4: reopenedIssues = count; break;
                        case 5: resolvedIssues = count; break;
                        case 6: closedIssues = count; break;
                        default: throw new IllegalArgumentException(String.format("Unknown issue status: %d", status));
                    }
                }

                // scrape issue priorities
                int blockerOpenIssues = 0;
                int criticalOpenIssues = 0;
                int majorOpenIssues = 0;
                int minorOpenIssues = 0;
                int trivialOpenIssues = 0;
                int optionalOpenIssues = 0;
                for(Iterator itIssuePriority = resultIssuePriority.iterator(); itIssuePriority.hasNext();) {
                    String valueIssuePriority = ((NodeInfo)itIssuePriority.next()).getStringValue();
                    log.debug(String.format("Scraped issue priority: \"%s\"", valueIssuePriority));

                    StringTokenizer tokenizerIssuePriority = new StringTokenizer(valueIssuePriority, "|");
                    int priority = Integer.parseInt(tokenizerIssuePriority.nextToken());
                    int count = Integer.parseInt(tokenizerIssuePriority.nextToken());
                    switch(priority) {
                        case 1: blockerOpenIssues = count; break;
                        case 2: criticalOpenIssues = count; break;
                        case 3: majorOpenIssues = count; break;
                        case 4: minorOpenIssues = count; break;
                        case 5: trivialOpenIssues = count; break;
                        // TODO what is 6?
                        case 7: optionalOpenIssues = count; break;
                        default: throw new IllegalArgumentException(String.format("Unknown issue priority: %d", priority));
                    }
                }

                // scrape issue assignees
                Map<String, Integer> openIssuesPerAssigneeMap = new HashMap<String, Integer>();
                for(Iterator itIssueAssignee = resultIssueAssignee.iterator(); itIssueAssignee.hasNext();) {
                    String valueIssueAssignee = ((NodeInfo)itIssueAssignee.next()).getStringValue();
                    log.debug(String.format("Scraped issue assignee: \"%s\"", valueIssueAssignee));

                    StringTokenizer tokenizerIssueAssignee = new StringTokenizer(valueIssueAssignee, "|");
                    String assignee = tokenizerIssueAssignee.nextToken();
                    int count = Integer.parseInt(tokenizerIssueAssignee.nextToken());
                    openIssuesPerAssigneeMap.put(assignee, count);
                }

                // generate charts
                String prefix = UrlUtils.removeUserInfoFromUrl(url);
                List<Map.Entry<String, Integer>> openIssuesPerAssignee = (List<Map.Entry<String, Integer>>)ChartUtils.countMapToSortedList(openIssuesPerAssigneeMap);

                ZoomableImageLocator openIssuesPerAssigneeChartLocator = generateOpenIssuesPerAssigneeChart(prefix, openIssuesPerAssignee);
                ZoomableImageLocator issuesPerStatusChartLocator = generateIssuePerStatusChart(prefix, openIssues, codingInProgressIssues, reopenedIssues, resolvedIssues, closedIssues);
                ZoomableImageLocator openIssuesPerPriorityChartLocator = generateIssuePerPriorityChart(prefix, blockerOpenIssues, criticalOpenIssues, majorOpenIssues, minorOpenIssues, trivialOpenIssues, optionalOpenIssues);

                // save
                JiraProject project = new JiraProject(name, prefix, projectKey, projectUrl, description, lead,
                    openIssues, codingInProgressIssues, reopenedIssues, resolvedIssues, closedIssues,
                    blockerOpenIssues, criticalOpenIssues, majorOpenIssues, minorOpenIssues, trivialOpenIssues, optionalOpenIssues,
                    (List<Map.Entry<String, Integer>>)trimList(openIssuesPerAssignee),
                    issuesPerStatusChartLocator, openIssuesPerPriorityChartLocator, openIssuesPerAssigneeChartLocator);
                log.debug(String.format("Adding %s...", project));
                projects.add(project);
            }
        } catch(Exception ex) {
            log.error("Unable to scrape", ex);
        }

        endStoreSession();

        return projects;
    }

    /** Connects to the given public or secure JIRA URL. */
    protected void connect(String url) throws MalformedURLException {
        // authenticate if project is secured
        String userInfo[] = UrlUtils.parseUserInfoFromUrl(url);
        if(userInfo != null) {
            String userName = userInfo[0];
            String password = userInfo[1];

            log.debug(String.format("Signing in as \"%s\" \"%s\"...", userName, password));
            // do nothing
        }
    }

    /** Disconnects from the given URL. */
    protected void disconnect(String url) {
        // do nothing
    }
}
